//asic_libusb.c: function of libusb IO;
/* sane - Scanner Access Now Easy.
	Copyright (C) 2005 Mustek.
	Originally maintained by Mustek
	Author:Roy 2005.5.24

	This file is part of the SANE package.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as
	published by the Free Software Foundation; either version 2 of the
	License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston,
	MA 02111-1307, USA.

	As a special exception, the authors of SANE give permission for
	additional uses of the libraries contained in this release of SANE.

	The exception is that, if you link a SANE library with other files
	to produce an executable, this does not by itself cause the
	resulting executable to be covered by the GNU General Public
	License.  Your use of that executable is in no way restricted on
	account of linking the SANE library code into it.

	This exception does not, however, invalidate any other reasons why
	the executable file might be covered by the GNU General Public
	License.

	If you submit changes to SANE to the maintainers to be included in
	a subsequent release, you agree by submitting the changes that
	those changes may be distributed with this exception intact.

	If you write modifications of your own for SANE, it is your choice
	whether to permit this exception to apply to your modifications.
	If you do not wish that, delete this exception notice.

	This file provides a generic USB interface.  */


#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <dirent.h>

#include <usb.h>
#include "asic_libusb.h"
#include "../include/debug.h"

typedef struct
{
	BOOL open;
	char* devname;
	int vendor;
	int product;
	int bulk_in_ep;
	int bulk_out_ep;
	int int_in_ep;
	int int_out_ep;
	int interface_nr;

	usb_dev_handle *libusb_handle;
	struct usb_device *libusb_device;
}
device_list_type;

/* total number of devices that can be opened at the same time */
#define MAX_DEVICES 100

/* per-device information, using the functions' parameters dn as index */
static device_list_type devices[MAX_DEVICES];

static int libusb_timeout = 30 * 1000;	/* 30 seconds */

static BOOL libusb_inited = FALSE;


static void print_buffer (const BYTE * buffer, int size)
{
	#define NUM_COLUMNS 16
	#define PRINT_BUFFER_SIZE (6 + NUM_COLUMNS * (3 + 1) + 1 + 1)
	char line_str [PRINT_BUFFER_SIZE];
	char *pp;
	int column;
	int line;
	
	memset (line_str, 0, PRINT_BUFFER_SIZE);
	

	for (line = 0; line < ((size + NUM_COLUMNS - 1) / NUM_COLUMNS); line++)
    {
		pp = line_str;
		sprintf (pp, "%04X: ", line * NUM_COLUMNS);
		pp += 6;
		for (column = 0; column < NUM_COLUMNS; column++)
		{
			if ((line * NUM_COLUMNS + column) < size)
				sprintf (pp, "%02X ", buffer [line * NUM_COLUMNS + column]);
			else
				sprintf (pp, "   ");
			pp += 3;
		}
		for (column = 0; column < NUM_COLUMNS; column++)
		{
			if ((line * NUM_COLUMNS + column) < size)
				sprintf (pp, "%c",
				(buffer [line * NUM_COLUMNS + column] < 127) && 
				(buffer [line * NUM_COLUMNS + column] > 31) ?
				buffer [line * NUM_COLUMNS + column] : '.');
			else
				sprintf (pp, ".");
			pp += 1;
		}
		XDBG(DEBUG_LIBUSB, "%s\n", line_str);
    }
}

STATUS	libusb_init (int vid,int pid, int *dn)
{
	XDBG(DEBUG_ASIC,"libusb_init: Enter");

	int vendor, product;
	char devname[1024];
	int devcount = 0;
	struct usb_bus *bus;
	struct usb_device *dev;
	
	if (libusb_inited)
		return STATUS_GOOD;
	libusb_inited = TRUE;
	
	memset (devices, 0, sizeof (devices));

	/* Check for devices using libusb */
	usb_init ();
	if (!usb_get_busses ())
    {
		usb_find_busses ();
		usb_find_devices ();
    }
	
	/* Check for the matching device */
	for (bus = usb_get_busses (); bus; bus = bus->next)
    {
		for (dev = bus->devices; dev; dev = dev->next)
		{
			int interface;
			BOOL found;
			
			if (!dev->config)
			{
				XDBG(DEBUG_LIBUSB,  "sanei_usb_init: device 0x%04x/0x%04x is not configured",
					dev->descriptor.idVendor, dev->descriptor.idProduct);
				continue;
			}
			if (dev->descriptor.idVendor == 0 || dev->descriptor.idProduct == 0)
			{
				XDBG(DEBUG_LIBUSB,  "libusb_init: device 0x%04x/0x%04x looks like a root hub",
					dev->descriptor.idVendor, dev->descriptor.idProduct);
				continue;
			}
			
			found = FALSE;
			for (interface = 0; interface < dev->config[0].bNumInterfaces && !found; interface++)
			{
				switch (dev->descriptor.bDeviceClass)
				{
				case USB_CLASS_VENDOR_SPEC:
					found = TRUE;
					break;
				case USB_CLASS_PER_INTERFACE:
					switch (dev->config[0].interface[interface].altsetting[0].bInterfaceClass)
					{
					case USB_CLASS_VENDOR_SPEC:
					case USB_CLASS_PER_INTERFACE:
					case 16:                /* data? */
						found = TRUE;
						break;
					}
					break;
				}
				if (!found)
					XDBG(DEBUG_LIBUSB,  "libusb_init: device 0x%04x/0x%04x, interface %d doesn't look like a "
					"scanner (%d/%d)", dev->descriptor.idVendor,
					dev->descriptor.idProduct, interface, dev->descriptor.bDeviceClass, 
					dev->config[0].interface[interface].altsetting[0].bInterfaceClass);
			}
			interface--;
			if (!found)
			{
				XDBG(DEBUG_LIBUSB,  "libusb_init: device 0x%04x/0x%04x: no suitable interfaces",
					dev->descriptor.idVendor, dev->descriptor.idProduct);
				continue;
			}
			
			devices[devcount].libusb_device = dev;
			snprintf (devname, sizeof (devname), "libusb:%s:%s",
				dev->bus->dirname, dev->filename);
			devices[devcount].devname = strdup (devname);
			if (!devices[devcount].devname)
				return STATUS_INVAL;
			devices[devcount].vendor = dev->descriptor.idVendor;
			devices[devcount].product = dev->descriptor.idProduct;
			devices[devcount].open = FALSE;
			devices[devcount].interface_nr = interface;

			XDBG(DEBUG_LIBUSB, "libusb_init: found libusb device (0x%04x/0x%04x) interface %d  at %s\n",
				dev->descriptor.idVendor, dev->descriptor.idProduct, interface, devname);

			if ((vid==dev->descriptor.idVendor)&&(pid==dev->descriptor.idProduct))
			{
				*dn = devcount;
				XDBG(DEBUG_LIBUSB,  "find device `%s' in dn=%d", devname,devcount);
				return STATUS_GOOD;
			}
      
 			devcount++;
			if (devcount >= MAX_DEVICES)
			{
				XDBG(DEBUG_LIBUSB,  "Can not find scanner in list");
				return STATUS_INVAL;
			}
		}
    }

	XDBG(DEBUG_LIBUSB,  "libusb_init: Exit");
	return STATUS_INVAL;
}


STATUS	libusb_open (int dn)
{
	XDBG(DEBUG_ASIC,"libusb_open: Enter");
	BOOL found = FALSE;
 	struct usb_device *dev;
	struct usb_interface_descriptor *interface;
	int result, num;
		
	
	if (devices[dn].open)
	{
		XDBG(DEBUG_LIBUSB, " device `%s' already open\n", devices[dn].devname);
		return STATUS_INVAL;
	}
	

	devices[dn].libusb_handle = usb_open (devices[dn].libusb_device);
	XDBG(DEBUG_LIBUSB, "after usb_open");
  
	if (!devices[dn].libusb_handle)
	{
		STATUS status = STATUS_INVAL;
		
		XDBG(DEBUG_LIBUSB, "libusb_open: can't open device `%s': %s\n",devices[dn].devname, strerror (errno));
		if (errno == EPERM)
		{
			XDBG(DEBUG_LIBUSB, "Make sure you run as root or set appropriate permissions\n");
			status = STATUS_INVAL;
		}
		else if (errno == EBUSY)
		{
			XDBG(DEBUG_LIBUSB,  "Maybe the kernel scanner driver claims the scanner's interface?\n");
			status = STATUS_DEVICE_BUSY;
		}
		return status;
	}
	
	dev = usb_device (devices[dn].libusb_handle);
	
	/* Set the configuration */
	if (!dev->config)
	{
		XDBG(DEBUG_LIBUSB,  "libusb_open: device `%s' not configured?\n", devices[dn].devname);
		return STATUS_INVAL;
	}
	if (dev->descriptor.bNumConfigurations > 1)
	{
		XDBG(DEBUG_LIBUSB,  "libusb_open: more than one configuration (%d), choosing first config (%d)\n",
			dev->descriptor.bNumConfigurations, 
			dev->config[0].bConfigurationValue);
	}
	result = usb_set_configuration (devices[dn].libusb_handle,
		dev->config[0].bConfigurationValue);
	if (result < 0)
	{
		STATUS status = STATUS_INVAL;
		
		XDBG(DEBUG_LIBUSB, "libusb_open: libusb complained: %s\n", usb_strerror ());
		if (errno == EPERM)
		{
			XDBG(DEBUG_LIBUSB,  "Make sure you run as root or set appropriate  permissions\n");
			status = STATUS_INVAL;
		}
		else if (errno == EBUSY)
		{
			XDBG(DEBUG_LIBUSB,  "Maybe the kernel scanner driver claims the  scanner's interface?\n");
			status = STATUS_DEVICE_BUSY;
		}
		usb_close (devices[dn].libusb_handle);
		return status;
	}
	
	/* Claim the interface */
	result = usb_claim_interface (devices[dn].libusb_handle, 
		devices[dn].interface_nr);
	if (result < 0)
	{
		STATUS status = STATUS_INVAL;
		
		XDBG(DEBUG_LIBUSB, "libusb_open: libusb complained: %s\n", usb_strerror ());
		if (errno == EPERM)
		{
			XDBG(DEBUG_LIBUSB, "Make sure you run as root or set appropriate  permissions\n");
			status = STATUS_INVAL;
		}
		else if (errno == EBUSY)
		{
			XDBG(DEBUG_LIBUSB,  "Maybe the kernel scanner driver claims the  scanner's interface?\n");
			status = STATUS_DEVICE_BUSY;
		}
		usb_close (devices[dn].libusb_handle);

		return status;
	}
	interface = &dev->config[0].interface->altsetting[0];
	
	/* Now we look for usable endpoints */
	for (num = 0; num < interface->bNumEndpoints; num++)
	{
		struct usb_endpoint_descriptor *endpoint;
		int address, direction, transfer_type;
		
		endpoint = &interface->endpoint[num];
		address = endpoint->bEndpointAddress & USB_ENDPOINT_ADDRESS_MASK;
		direction = endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK;
		transfer_type = endpoint->bmAttributes & USB_ENDPOINT_TYPE_MASK;
		
		/* save the endpoints we need later */
		if (transfer_type == USB_ENDPOINT_TYPE_INTERRUPT)
		{
			XDBG(DEBUG_LIBUSB,  "libusb_open: found interupt-%s endpoint (address %d)\n",
				direction ? "in" : "out", address);
			if (direction)	/* in */
			{
				if (devices[dn].int_in_ep)
					XDBG(DEBUG_LIBUSB,  "libusb_open: we already have a int-in endpoint "
					"(address: %d), ignoring the new one\n",
					devices[dn].int_in_ep);
				else
					devices[dn].int_in_ep = endpoint->bEndpointAddress;
			}
			else
				if (devices[dn].int_out_ep)
					XDBG(DEBUG_LIBUSB,  "libusb_open: we already have a int-out endpoint "
					"(address: %d), ignoring the new one\n",
					devices[dn].int_out_ep);
				else
					devices[dn].int_out_ep = endpoint->bEndpointAddress;
		}
		else if (transfer_type == USB_ENDPOINT_TYPE_BULK)
		{
			XDBG(DEBUG_LIBUSB,  "libusb_open: found bulk-%s endpoint (address %d)\n",
				direction ? "in" : "out", address);
			if (direction)	/* in */
			{
				if (devices[dn].bulk_in_ep)
					XDBG(DEBUG_LIBUSB, "libusb_open: we already have a bulk-in endpoint "
					"(address: %d), ignoring the new one\n",
					devices[dn].bulk_in_ep);
				else
					devices[dn].bulk_in_ep = endpoint->bEndpointAddress;
			}
			else
			{
				if (devices[dn].bulk_out_ep)
					XDBG(DEBUG_LIBUSB, "libusb_open: we already have a bulk-out endpoint "
					"(address: %d), ignoring the new one\n",
					devices[dn].bulk_out_ep);
				else
					devices[dn].bulk_out_ep = endpoint->bEndpointAddress;
			}
		}
		/* ignore currently unsupported endpoints */
		else {
			XDBG(DEBUG_LIBUSB,  "libusb_open: ignoring %s-%s endpoint "
				"(address: %d)\n",
				transfer_type == USB_ENDPOINT_TYPE_CONTROL ? "control" :
			transfer_type == USB_ENDPOINT_TYPE_ISOCHRONOUS
				? "isochronous" : "interrupt",
				direction ? "in" : "out", address);
			continue;
		}
	}
	
	devices[dn].open = TRUE;
	
	XDBG(DEBUG_LIBUSB, "libusb_open: opened usb device `%s' (*dn=%d)\n",devices[dn].devname, dn);
	XDBG(DEBUG_ASIC,"libusb_open: Exit");
	return STATUS_GOOD;
}

void libusb_close (int dn)
{
	XDBG(DEBUG_ASIC,"libusb_close: Enter");
	XDBG(DEBUG_LIBUSB, "libusb_close: closing device %d\n", dn);
	if (dn >= MAX_DEVICES || dn < 0)
    {
		XDBG(DEBUG_LIBUSB,  "libusb_close: dn >= MAX_DEVICES || dn < 0\n");
		return;
    }
	if (!devices[dn].open)
    {
		XDBG(DEBUG_LIBUSB, "libusb_close: device %d already closed or never opened\n",
			dn);
		return;
    }

	#if 0	
	/* Should only be done in case of a stall */
	usb_clear_halt (devices[dn].libusb_handle, devices[dn].bulk_in_ep);
	usb_clear_halt (devices[dn].libusb_handle, devices[dn].bulk_out_ep);
	/* be careful, we don't know if we are in DATA0 stage now */
	usb_resetep(devices[dn].libusb_handle, devices[dn].bulk_in_ep);
	usb_resetep(devices[dn].libusb_handle, devices[dn].bulk_out_ep);
	#endif

	usb_release_interface (devices[dn].libusb_handle, devices[dn].interface_nr);
	usb_close (devices[dn].libusb_handle);

	devices[dn].open = FALSE;
	XDBG(DEBUG_ASIC,"libusb_close: Exit");
	return;
}

STATUS	libusb_read_bulk (int dn, BYTE * buffer, size_t * size)
{
	XDBG(DEBUG_LIBUSB,"libusb_read_bulk: Enter");
	ssize_t read_size = 0;
	
	if (!size)
    {
		XDBG(DEBUG_LIBUSB, "libusb_read_bulk: size == NULL\n");
		return STATUS_INVAL;
    }
	
	if (dn >= MAX_DEVICES || dn < 0)
    {
		XDBG(DEBUG_LIBUSB,  "libusb_read_bulk: dn >= MAX_DEVICES || dn < 0\n");
		return STATUS_INVAL;
    }
	XDBG(DEBUG_LIBUSB,  "libusb_read_bulk: trying to read %lu bytes\n",
		(unsigned long) *size);
	
	if (devices[dn].bulk_in_ep)
		read_size = usb_bulk_read (devices[dn].libusb_handle,
		devices[dn].bulk_in_ep, (char *) buffer,
		(int) *size, libusb_timeout);
	else
	{
		XDBG(DEBUG_LIBUSB,  "libusb_read_bulk: can't read without a bulk-in "
			"endpoint\n");
		return STATUS_INVAL;
	}

	
	if (read_size < 0)
    {
		XDBG(DEBUG_LIBUSB, "libusb_read_bulk: read failed: %s\n", strerror (errno));

			usb_clear_halt (devices[dn].libusb_handle, devices[dn].bulk_in_ep);
		*size = 0;
		return STATUS_IO_ERROR;
    }
	if (read_size == 0)
    {
		XDBG(DEBUG_LIBUSB,  "libusb_read_bulk: read returned EOF\n");
		*size = 0;
		return STATUS_EOF;
    }

	/*	print_buffer (buffer, read_size);*/

	XDBG(DEBUG_LIBUSB, "libusb_read_bulk: wanted %lu bytes, got %ld bytes\n",
		(unsigned long) *size, (unsigned long) read_size);
	*size = read_size;
    
	XDBG(DEBUG_LIBUSB,"libusb_read_bulk: Exit");
	return STATUS_GOOD;
}


STATUS	libusb_write_bulk (int dn, const BYTE * buffer, size_t * size)
{
	XDBG(DEBUG_LIBUSB,"libusb_write_bulk: Enter");
	ssize_t write_size = 0;
	
	if (!size)
    {
		XDBG(DEBUG_LIBUSB,  "libusb_write_bulk: size == NULL\n");
		return STATUS_INVAL;
    }
	
	if (dn >= MAX_DEVICES || dn < 0)
    {
		XDBG(DEBUG_LIBUSB,  "libusb_write_bulk: dn >= MAX_DEVICES || dn < 0\n");
		return STATUS_INVAL;
    }
	XDBG(DEBUG_LIBUSB,  "libusb_write_bulk: trying to write %lu bytes\n",
		(unsigned long) *size);

	/*	print_buffer (buffer, *size);*/
	
		if (devices[dn].bulk_out_ep)
			write_size = usb_bulk_write (devices[dn].libusb_handle,
			devices[dn].bulk_out_ep,
			(const char *) buffer,
			(int) *size, libusb_timeout);
		else
		{
			XDBG(DEBUG_LIBUSB, "libusb_write_bulk: can't write without a bulk-out endpoint\n");
			return STATUS_INVAL;
		}

	
	if (write_size < 0)
    {
		XDBG(DEBUG_LIBUSB, "libusb_write_bulk: write failed: %s\n", strerror (errno));
		*size = 0;
		usb_clear_halt (devices[dn].libusb_handle, devices[dn].bulk_out_ep);
		return STATUS_IO_ERROR;
    }
	XDBG(DEBUG_LIBUSB,  "libusb_write_bulk: wanted %lu bytes, wrote %ld bytes\n",
		(unsigned long) *size, (unsigned long) write_size);
	*size = write_size;

	XDBG(DEBUG_LIBUSB,"libusb_write_bulk: Exit");
	return STATUS_GOOD;
}

STATUS	libusb_control_msg (int dn, int rtype, int req,
					   int value, int index, int len,
					   BYTE * data)
{
	XDBG(DEBUG_LIBUSB,"libusb_control_msg: Enter");
	
	if (dn >= MAX_DEVICES || dn < 0)
    {
		XDBG(DEBUG_LIBUSB,  "libusb_control_msg: dn >= MAX_DEVICES || dn < 0\n");
		return STATUS_INVAL;
    }
	
	XDBG(DEBUG_LIBUSB,  "libusb_control_msg: rtype = 0x%02x, req = %d, value = %d, "
		"index = %d, len = %d\n", rtype, req, value, index, len);
	
	
	int result;
	
	result = usb_control_msg (devices[dn].libusb_handle, rtype, req,
		value, index, (char *) data, len,
		libusb_timeout);
	if (result < 0)
	{
		XDBG(DEBUG_LIBUSB,  "libusb_control_msg: libusb complained: %s\n",
			usb_strerror ());
		return STATUS_INVAL;
	}
	
	XDBG(DEBUG_LIBUSB,"libusb_control_msg: Exit");
	return STATUS_GOOD;
}
